'''
Click base group function.
'''

import click

from posce.items import Book

SETTINGS = {
    'help_option_names': ['-h', '--help'],
}

@click.group(context_settings=SETTINGS)
@click.option('--dire', envvar='POSCE_DIR', hidden=True)
@click.option('--ext',  envvar='POSCE_EXT', hidden=True)
@click.pass_context
def group(ctx, dire, ext):
    '''
    Posce: a note-taking toolkit for your command line.
    See github.com/posce/posce for help and issues.
    '''

    ctx.obj = Book(dire, ext)



'''
Package definition for 'posce'.
'''

VERSION_DATE   = '2020-03-18'
VERSION_NUMBER = '0.1.1'
VERSION_STRING = f'Posce version {VERSION_NUMBER} ({VERSION_DATE}).'

from posce import comms
from posce import items
from posce import tools



'''
Class definition for 'Book'.
'''

from posce.items.note import Note
from posce            import tools

class Book:
    '''
    A directory of plaintext note files.
    '''

    __slots__ = ['dire', 'ext', 'notes']

    def __init__(self, dire, ext):
        '''
        Initialise a new Book.
        '''

        self.dire  = str(dire)
        self.ext   = str(ext)
        self.notes = {
            note.name: note for note in
            map(Note, tools.file.find(self.dire, f'*.{self.ext}'))
        }

    def __contains__(self, name):
        '''
        Return True if the Book contains a Note name.
        '''

        return name in self.notes.keys()

    def __eq__(self, book):
        '''
        Return True if the Book is equal to another Book.
        '''

        return all([
            isinstance(book, Book),
            self.dire == getattr(book, 'dire', None),
            self.ext  == getattr(book, 'ext',  None),
        ])

    def __getitem__(self, name):
        '''
        Return a Note in the Book by name.
        '''

        return self.notes[name]

    def __hash__(self):
        '''
        Return the Book's unique hash code.
        '''

        return hash(self.dire + self.ext)

    def __iter__(self):
        '''
        Yield each Note in the Book.
        '''

        yield from self.notes.values()

    def __len__(self):
        '''
        Return the number of Notes in the Book.
        '''

        return len(self.notes)

    def __repr__(self):
        '''
        Return the Book as a code-representative string.
        '''

        return f'Book({self.dire!r}, {self.ext!r})'

    def create(self, name, string):
        '''
        Create and return a new Note in the Book.
        '''

        dest = tools.path.join(self.dire, f'{name}.{self.ext}')
        tools.file.create(dest, string)
        self.notes[name] = Note(dest)
        return self.notes[name]

    def exists(self):
        '''
        Return True if the Book's directory exists.
        '''

        return tools.file.exists(self.dire)

    def filter(self, func):
        '''
        Yield each Note in the Book passing a filter function.
        '''

        for note in self.notes.values():
            if func(note):
                yield note

    def match(self, pattern):
        '''
        Yield each Note with a name matching a glob pattern.
        '''

        yield from self.filter(lambda note: note.match(pattern))

    def search(self, pattern):
        '''
        Yield each Note with contents matching a regular expression.
        '''

        yield from self.filter(lambda note: note.search(pattern))



'''
Class definition for 'Note'.
'''

import fnmatch
import re

from posce import tools

class Note:
    '''
    A single plaintext note file.
    '''

    __slots__ = ['ext', 'name', 'path']

    def __init__(self, path):
        '''
        Initialise a new Note.
        '''

        self.path = str(path)
        self.ext  = tools.path.ext(self.path)
        self.name = tools.path.name(self.path)

    def __contains__(self, string):
        '''
        Return True if the Note contains a string.
        '''

        return string in self.read()

    def __eq__(self, note):
        '''
        Return True if the Note is equal to another Note.
        '''

        return all([
            isinstance(note, Note),
            self.path == getattr(note, 'path', None),
        ])

    def __hash__(self):
        '''
        Return the Note's unique hash code.
        '''

        return hash(self.path)

    def __iter__(self):
        '''
        Yield the Note's contents as lines.
        '''

        yield from self.read().splitlines(keepends=True)

    def __len__(self):
        '''
        Return the length of the Note's contents.
        '''

        return len(self.read())

    def __repr__(self):
        '''
        Return the Note as a code-representative string.
        '''

        return f'Note({self.path!r})'

    def append(self, string, *, sep='\n'):
        '''
        Append a string to the Note's contents.
        '''

        tools.file.append(self.path, string, sep=sep)

    def exists(self):
        '''
        Return True if the Note's path exists.
        '''

        return tools.file.exists(self.path)

    def match(self, pattern):
        '''
        Return True if the Note's name matches a glob pattern.
        '''

        return fnmatch.fnmatch(self.name, pattern)

    def read(self):
        '''
        Return the contents of the Note as a string.
        '''

        return tools.file.read(self.path)

    def search(self, pattern):
        '''
        Return True if the Note's contents matches a regular expression.
        '''

        return re.search(pattern, self.read())

    def write(self, string):
        '''
        Overwrite the Note's contents with a string.
        '''

        tools.file.write(self.path, string)



'''
Click command function 'list'.
'''

import click

from posce.comms.base import group

@group.command()
@click.argument('glob',
    default = '*',
)
@click.option('-r', '--reverse',
    help    = 'Reverse sort order.',
    is_flag = True,
)
@click.option('-s', '--sort',
    help    = 'Sort notes by element.',
    default = 'name',
    type    = click.Choice(['name', 'size']),
)
@click.pass_obj
def list(book, glob, sort, reverse):
    '''
    List all notes.
    '''

    key = {
        'name': lambda note: note.name,
        'size': lambda note: len(note),
    }[sort]

    for note in sorted(book.match(glob), key=key, reverse=reverse):
        click.echo(note.name)



'''
PyPi setup script.
'''

from setuptools import find_packages, setup

from posce import VERSION_NUMBER

setup(
    # Basic information.
    name         = 'posce',
    version      = VERSION_NUMBER,
    keywords     = 'cli notes note-taking',
    description  = 'A note-taking toolkit for your command line.',
    author       = 'Stephen Malone',
    author_email = 'mail@posce.org',

    # Project description.
    long_description = open('readme.md').read(),
    long_description_content_type = 'text/markdown',

    # Package specifications.
    packages         = find_packages(exclude=['*tests*']),
    python_requires  = '>=3.8.0',
    install_requires = ['click>=7.1.1'],

    # Console executables.
    entry_points = {
        'console_scripts': ['posce=posce.__main__:main'],
    },

    # Project URLs.
    project_urls = {
        'Homepage': 'https://github.com/posce/posce',
        'Issues':   'https://github.com/posce/posce/issues',
    },

    # Project classifiers.
    classifiers = [
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.8',
        'Topic :: Office/Business :: News/Diary',
        'Topic :: Office/Business',
    ],
)
